'''
Copyright (C) 2024 1P2D
https://1p2d.com
me@1p2d.com

Created by 1P2D - Felix Pierron

This file is part of the Shading + pie.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, see <https://www.gnu.org/licenses>.
'''

bl_info = {
    "name": "HDRI Rotator",
    "description": "Rotate your HDRI and shadows from the viewport with your mouse",
    "author": "1P2D <me@1p2d.com>",
    "version": (1, 0, 0),
    "blender": (4, 2, 0),
    "category": "Interface",
    "location": "3D View",
    "doc_url": "https://www.1p2d.com/"
}

import bpy
import math
import mathutils

class HDRI_Rotator(bpy.types.Operator):
    """Rotate HDRI or Light Direction with ALT + LMB drag"""
    bl_idname = "view3d.hdri_rotator"
    bl_label = "HDRI Rotator"
    bl_options = {'REGISTER', 'GRAB_CURSOR', 'BLOCKING'}

    def invoke(self, context, event):
        self.start_x = event.mouse_x
        self.start_y = event.mouse_y

        # Store the window, area, space, and scene to use during modal execution
        self.window = context.window
        self.area = context.area
        self.space = context.space_data
        self.scene = context.scene

        # Ensure we are in a 3D Viewport
        if self.area.type != 'VIEW_3D':
            self.report({'WARNING'}, "Operator must be run in a 3D Viewport")
            return {'CANCELLED'}

        self.shading = self.space.shading
        self.shading_type = self.shading.type

        if self.shading_type == 'SOLID' and self.shading.show_shadows:
            # Solid shading mode with shadows enabled; rotate light direction
            self.mode = 'LIGHT_DIRECTION'
            self.start_light_direction = self.scene.display.light_direction.copy()
            # Set cursor icon for light direction rotation
            self.window.cursor_modal_set('SCROLL_XY')
        else:
            # Other shading modes; rotate HDRI
            # Get the use_scene_world setting from the shading settings
            self.use_scene_world = self.shading.use_scene_world_render
            if self.use_scene_world:
                # For custom HDRI
                self.rotation_input = self.find_rotation_input()
                if self.rotation_input is not None:
                    self.start_rotation = self.rotation_input.default_value
                    self.mode = 'CUSTOM_HDRI_ROTATION'
                else:
                    # Try to find a Mapping node connected to an Image Texture node
                    self.mapping_node = self.find_mapping_node()
                    if self.mapping_node is None:
                        self.report({'WARNING'}, "No 'Rotation' input or Mapping node found in World nodes")
                        self.window.cursor_modal_restore()
                        return {'CANCELLED'}
                    self.start_rotation = self.mapping_node.inputs['Rotation'].default_value[2]  # Z rotation
                    self.mode = 'CUSTOM_HDRI_ROTATION'
            else:
                # For built-in HDRI
                self.start_rotation = self.get_builtin_hdr_rotation()
                self.mode = 'HDRI_ROTATION'
            # Set cursor icon for HDRI rotation
            self.window.cursor_modal_set('SCROLL_X')

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            delta_x = event.mouse_x - self.start_x
            delta_y = event.mouse_y - self.start_y

            if self.mode == 'LIGHT_DIRECTION':
                # Rotate light direction based on mouse movement
                sensitivity = 0.005  # Adjust sensitivity as needed
                rot_x = delta_y * sensitivity  # Vertical movement rotates around X axis
                rot_z = -delta_x * sensitivity  # Horizontal movement rotates around Z axis

                new_light_direction = self.calculate_new_direction(
                    self.start_light_direction,
                    rot_x,
                    rot_z
                )
                self.scene.display.light_direction = new_light_direction
                if self.area:
                    self.area.tag_redraw()
            elif self.mode == 'CUSTOM_HDRI_ROTATION':
                # Rotate custom HDRI based on mouse movement
                delta_rotation = delta_x * 0.1  # Adjust sensitivity as needed
                new_rotation = self.start_rotation + delta_rotation
                if self.rotation_input is not None:
                    self.set_custom_hdri_rotation(new_rotation)
                elif self.mapping_node is not None:
                    self.set_mapping_node_rotation(new_rotation)
            elif self.mode == 'HDRI_ROTATION':
                # Rotate built-in HDRI based on mouse movement
                delta_rotation = delta_x * 0.1
                new_rotation = self.start_rotation + delta_rotation
                self.set_builtin_hdr_rotation(new_rotation)
            return {'RUNNING_MODAL'}
        elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':
            # Restore cursor
            self.window.cursor_modal_restore()
            return {'FINISHED'}
        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            # Restore cursor
            self.window.cursor_modal_restore()
            if self.mode == 'LIGHT_DIRECTION':
                # Restore original light direction
                self.scene.display.light_direction = self.start_light_direction
                if self.area:
                    self.area.tag_redraw()
            return {'CANCELLED'}
        return {'RUNNING_MODAL'}

    def find_rotation_input(self, node_tree=None, visited=None):
        if visited is None:
            visited = set()
        if node_tree is None:
            world = self.scene.world
            if world is None or not world.use_nodes:
                return None
            node_tree = world.node_tree

        if node_tree in visited:
            return None
        visited.add(node_tree)

        for node in node_tree.nodes:
            if node.type == 'GROUP':
                if 'Rotation' in node.inputs:
                    return node.inputs['Rotation']
                else:
                    # Recursively check inside the node group's node tree
                    rotation_input = self.find_rotation_input(node.node_tree, visited)
                    if rotation_input:
                        return rotation_input
        return None

    def find_mapping_node(self):
        world = self.scene.world
        node_tree = world.node_tree

        for node in node_tree.nodes:
            if node.type == 'MAPPING':
                # Check if this Mapping node is connected to an Image Texture node
                for output in node.outputs:
                    for link in output.links:
                        next_node = link.to_node
                        if next_node.type in {'TEX_ENVIRONMENT', 'TEX_IMAGE'}:
                            return node
        return None

    def set_custom_hdri_rotation(self, rotation_value):
        # Set the new rotation value
        self.rotation_input.default_value = rotation_value
        # Force update of the scene to ensure the change takes effect
        self.scene.view_layers[0].update()
        # Redraw the viewport
        if self.area:
            self.area.tag_redraw()

    def set_mapping_node_rotation(self, rotation_value):
        # Set the Z rotation of the Mapping node
        rotation = self.mapping_node.inputs['Rotation'].default_value.copy()
        rotation[2] = rotation_value  # Z rotation
        self.mapping_node.inputs['Rotation'].default_value = rotation
        # Force update
        self.scene.view_layers[0].update()
        if self.area:
            self.area.tag_redraw()

    def get_builtin_hdr_rotation(self):
        # Get the built-in HDRI rotation in degrees
        rotation_rad = self.shading.studiolight_rotate_z
        rotation_deg = math.degrees(rotation_rad)
        return rotation_deg

    def set_builtin_hdr_rotation(self, rotation_value):
        # Set the built-in HDRI rotation
        rotation_rad = math.radians(rotation_value)
        # Wrap rotation to -pi to pi
        rotation_rad_wrapped = (rotation_rad + math.pi) % (2 * math.pi) - math.pi
        self.shading.studiolight_rotate_z = rotation_rad_wrapped
        if self.area:
            self.area.tag_redraw()

    def calculate_new_direction(self, start_direction, rot_x, rot_z):
        # Rotate the start_direction vector around X and Z axes
        rot_x_mat = mathutils.Matrix.Rotation(rot_x, 4, 'X')
        rot_z_mat = mathutils.Matrix.Rotation(rot_z, 4, 'Z')

        # Apply rotations
        new_direction = rot_z_mat @ rot_x_mat @ start_direction

        # Normalize the vector
        if new_direction.length_squared != 0:
            new_direction.normalize()
        else:
            new_direction = mathutils.Vector((0.0, 0.0, -1.0))  # Default direction

        # Ensure the Z component is negative (light from above)
        if new_direction.z > 0:
            new_direction.z = -new_direction.z

        return new_direction

    def execute(self, context):
        return {'FINISHED'}

    def cancel(self, context):
        # Restore cursor on cancel
        self.window.cursor_modal_restore()
        return {'CANCELLED'}

def register():
    bpy.utils.register_class(HDRI_Rotator)
    
    # Set up keymap for ALT + LMB in the 3D View
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.new(name='3D View', space_type='VIEW_3D')
        kmi = km.keymap_items.new(HDRI_Rotator.bl_idname, 'LEFTMOUSE', 'PRESS', alt=True)
        kmi.active = True

def unregister():
    # Remove keymap
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.get('3D View')
        if km:
            for kmi in km.keymap_items:
                if kmi.idname == HDRI_Rotator.bl_idname:
                    km.keymap_items.remove(kmi)
                    break
    bpy.utils.unregister_class(HDRI_Rotator)

if __name__ == "__main__":
    register()
